# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import re
import logging
import time
import datetime

import pydub # https://github.com/jiaaro/pydub
from google.cloud import storage
from google.cloud.exceptions import NotFound
from google.cloud import speech
from google.cloud.speech import enums
from google.cloud.speech import types

from .path import Path, TempPath
from .utils import Time, String


logger = logging.getLogger(__name__)


class SpeechPath(Path):
    regex = re.compile(r"\.(?:mp3|wav|flac)$", re.I)

    def is_type(self):
        return True if self.regex.search(self) else False


class Speech(object):
    """Wrapper around Google's speech api

    https://cloud.google.com/speech
    https://cloud.google.com/speech/docs/reference/libraries#client-libraries-usage-python
    https://github.com/GoogleCloudPlatform/google-cloud-python/blob/master/speech/
    https://cloud.google.com/speech/support
    https://cloud.google.com/speech/docs/best-practices
    """

    TIMEOUT = 10800

    @property
    def text(self):
        return self._text

    @property
    def lines(self):
        return self.text.splitlines(False)

    @property
    def words(self):
        return self._words

    def __init__(self, path, lang):
        self.path = path
        self.lang = lang

    def _convert(self, path, start, stop):
        audio_format = "ogg"
        # https://github.com/jiaaro/pydub/blob/master/pydub/audio_segment.py
        audio = pydub.AudioSegment.from_file(path)
        if bool(start) or bool(stop):
            start = Time(start)
            stop = Time(stop)
            if start and not stop:
                logger.info("Starting audio at {}".format(start))
                audio = audio[start.total_ms():]
            elif stop and not start:
                logger.info("Stopping audio at {}".format(stop))
                audio = audio[:stop.total_ms():]
            else:
                logger.info("Slicing audio {} - {}".format(start, stop))
                audio = audio[start.total_ms():stop.total_ms()]

            tp = TempPath(
                "{}-{}-{}".format(
                    path.name,
                    start.total_seconds(),
                    stop.total_seconds()
                ),
                audio_format
            )

        else:
            tp = TempPath(path.name, audio_format)

        logger.info("Converting {} to {} at {}".format(path, audio_format, tp))

        # https://cloud.google.com/speech/docs/basics#audio-encodings
        # https://superuser.com/questions/516806/how-to-encode-audio-with-opus-codec
        #audio.export(tp, format=audio_format, codec="opus", parameters=["-strict", "-2"])
        ogg_audio = pydub.AudioSegment.from_file(
            audio.export(tp, format=audio_format, codec="libopus", parameters=["-ac", "1"])
        )

        return tp, ogg_audio

    def _upload(self, path):
        # https://cloud.google.com/python/getting-started/using-cloud-storage
        # https://github.com/GoogleCloudPlatform/google-cloud-python/
        # https://cloud.google.com/storage/docs/object-basics
        # https://googlecloudplatform.github.io/google-cloud-python/latest/storage/client.html
        storage_client = storage.Client()
        bucket_name = storage_client.project
        logger.info("Uploading {} to Google cloud storage bucket {}".format(path, bucket_name))

        try:
            bucket = storage_client.get_bucket(bucket_name)

        except NotFound:
            bucket = storage_client.create_bucket(bucket_name)

        filename = path.safename
        blob = bucket.blob(filename)
        blob.upload_from_string(path.contents(), path.mimetype)
        #public_url = blob.public_url
        # super convenient google doesn't bother with a .uri property for a
        # thing that seems to be pretty commonly needed
        # https://stackoverflow.com/a/25373496/5006
        uri = "gs://{}/{}".format(bucket_name, blob.name)
        logger.info("Google cloud storage uri {}".format(uri))
        return uri, blob

    def _transcribe(self, uri, audio):
        client = speech.SpeechClient()

        # NOTE -- according to https://cloud.google.com/speech/quotas  we are
        # limited to ~180 Minutes
        recog = types.RecognitionAudio(uri=uri)
        # https://cloud.google.com/speech/reference/rest/v1/RecognitionConfig
        config = types.RecognitionConfig(
            encoding=enums.RecognitionConfig.AudioEncoding.OGG_OPUS,
            sample_rate_hertz=audio.frame_rate,
            language_code=self.lang,
            enable_word_time_offsets=True
        )

        # https://cloud.google.com/speech/docs/async-recognize
        # https://googlecloudplatform.github.io/google-cloud-python/latest/speech/index.html#asynchronous-recognition
        operation = client.long_running_recognize(config, recog)

        start = time.time()
        logger.info('Beginning transcription')
        response = operation.result(timeout=self.TIMEOUT)
        stop = time.time()
        logger.info("Transcription finished in {}".format(Time(stop - start)))
        return response

    def scan(self, start=0, stop=0):
        path = self.path
        path, audio = self._convert(path, start, stop)
        uri, blob = self._upload(path)
        self.response = self._transcribe(uri, audio)
        logger.info("Deleting file from Google cloud storage {}".format(uri))
        blob.delete() # https://cloud.google.com/storage/docs/object-basics
        return self.response

    def __iter__(self):
        for result in self.response.results:
            alternative = result.alternatives[0]
            text = String(alternative.transcript).flow()

            # https://cloud.google.com/speech/docs/async-time-offsets
            word_info = alternative.words[0]
            # TODO -- we have a nanos property also, do we care about
            # nanoseconds?
            start_time = Time(word_info.start_time.seconds)
            yield start_time, text

#         for result in response.results:
#             # The first alternative is the most likely one for this portion.
#             print(result.alternatives[0].transcript)
#             #print('Confidence: {}'.format(result.alternatives[0].confidence))




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import os
import re

from google.cloud import vision

from .path import Path
from .utils import String


class ImagePath(Path):
    regex = re.compile(r"\.(?:jpe?g|gif|bmp|png|ico|tiff)$", re.I)

    def is_type(self):
        return True if self.regex.search(self) else False


class OCR(object):
    """
    https://googlecloudplatform.github.io/google-cloud-python/stable/vision-usage.html#text-detection
    https://github.com/GoogleCloudPlatform/google-cloud-python/tree/master/vision

    https://cloud.google.com/vision/docs/reference/libraries
    https://googlecloudplatform.github.io/google-cloud-python/latest/vision/gapic/v1/api.html
    https://googlecloudplatform.github.io/google-cloud-python/latest/vision/index.html#api-reference
    """
    @property
    def unformatted(self):
        """Return the original text (as it was OCR'd) without any formatting"""
        return String(self._text)

    @property
    def text(self):
        return self.unformatted.flow()

    @property
    def words(self):
        return self._words

    def __init__(self, path):
        self.path = path

    def scan(self):
        client = vision.ImageAnnotatorClient()
        image = vision.types.Image(content=self.path.contents())
        #image = client.image(content=self.path.contents())
        #self.response = image.document_text_detection(self.path.contents())
        self.response = client.document_text_detection(image)
        #pout.v(self.response)

        self._text = ""
        self._words = None
        if self.response:
            self._text = self.response.text_annotations[0].description
            self._words = (w.description for w in self.response.text_annotations[1:])

        return self.response




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import os
import tempfile
import re
import mimetypes


class Path(str):
    @property
    def mimetype(self):
        mt = mimetypes.guess_type(self)[0]
        if mt is None:
            mt = ""
        return mt

    @property
    def name(self):
        root, ext = os.path.splitext(self.basename)
        return root

    @property
    def basename(self):
        return os.path.basename(self)

    @property
    def safename(self):
        name = self.basename
        for c in "({[]})":
            name = name.replace(c, "")
        #name = re.strip(r"[\(\)\{\}\[\]]", "", name)
        name = name.replace(" ", "_")
        return name

    @property
    def ext(self):
        root, ext = os.path.splitext(self)
        return ext.lstrip(".")

    def __new__(cls, path):
        path = os.path.abspath(os.path.expanduser(path))
        return super(Path, cls).__new__(cls, path)

    def stream(self, mode="r"):
        return open(self, mode=mode)

    def is_file(self):
        return os.path.isfile(self)

    def is_dir(self):
        return os.path.isdir(self)

    def contents(self):
        if not self.is_file():
            raise RuntimeError("Trying to get the contents of a non-file")

        mode = "rb"
        with open(self, mode=mode) as fp:
            return fp.read()

    def is_type(self):
        """Override this in child classes to decide if this file should be returned
        in the iterator"""
        return True

    def __iter__(self):
        if self.is_file():
            if self.is_type():
                yield self

        elif self.is_dir():
            for basedir, directories, files in os.walk(self, topdown=True):
                for basename in files:
                    path = os.path.join(basedir, basename)
                    instance = type(self)(path)
                    if instance.is_type():
                        yield instance

        else:
            raise RuntimeError("Trying to iterate a non-directory or non-file")


class TempPath(Path):
    """The idea of this class is to make it easy to create a filepath in the temp
    directory

    :Example:
        tp = TempPath("foo", "txt") # full path will be something like: /tmp/foo.txt
    """
    def __new__(cls, basename, ext=""):
        basedir = tempfile.gettempdir()
        if ext:
            basename = "{}.{}".format(basename, ext.lstrip("."))
        path = os.path.join(basedir, basename)
        return super(TempPath, cls).__new__(cls, path)


class Paths(object):
    def __init__(self, paths, path_cls=Path):
        self.paths = paths
        self.path_cls = path_cls

    def __iter__(self):
        for path in self.paths:
            p = self.path_cls(path)
            for f in p:
                yield p





#!/usr/bin/env python
# http://docs.python.org/distutils/setupscript.html
# http://docs.python.org/2/distutils/examples.html
from setuptools import setup, find_packages
import re
import os
from codecs import open


name = "transcribe"

kwargs = {"name": name}

def read(path):
    if os.path.isfile(path):
        with open(path, encoding='utf-8') as f:
            return f.read()
    return ""


vpath = os.path.join(name, "__init__.py")
if os.path.isfile(vpath):
    kwargs["packages"] = find_packages(exclude=["tests", "tests.*", "examples"])
else:
    vpath = "{}.py".format(name)
    kwargs["py_modules"] = [name]
kwargs["version"] = re.search(r"^__version__\s*=\s*[\'\"]([^\'\"]+)", read(vpath), flags=re.I | re.M).group(1)
kwargs["long_description"] = read('README.rst')

tests_modules = []
install_modules = [
    "captain",
    "pydub",
    "google-cloud-vision",
    "google-cloud-storage",
    "google-cloud-speech",
]

setup(
    description='Convert images or audio files to plain text on the command line',
    keywords="ocr transcription speech-to-text speech-recognition command-line-tool",
    author='Jay Marcyes',
    author_email='jay@marcyes.com',
    url='http://github.com/Jaymon/{}'.format(name),
    #py_modules=[name], # files
    license="MIT",
    install_requires=install_modules,
    tests_require=tests_modules,
    classifiers=[ # https://pypi.python.org/pypi?:action=list_classifiers
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'Topic :: Text Processing',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
    ],
    entry_points = {
        'console_scripts': [
            '{} = {}.__main__:console'.format(name, name),
        ],
    },
    **kwargs
)




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import logging
import sys

from captain import exit, echo
from captain.decorators import arg

from transcribe import __version__
from transcribe.path import Paths
from transcribe.image import ImagePath, OCR
from transcribe.speech import SpeechPath, Speech
from transcribe.utils import Time


logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
log_handler = logging.StreamHandler(stream=sys.stderr)
log_formatter = logging.Formatter('[%(levelname).1s] %(message)s')
log_handler.setFormatter(log_formatter)
logger.addHandler(log_handler)


@arg("paths", nargs="+", default=[], help="path to an image or directory containing images")
@arg("--no-flow",
     dest="noflow",
     action="store_true",
     help="text will not be reformatted to remove unnecessary newlines",
)
@arg("--words", "-w", action="store_true", help="print all the words, one on each line")
def main_ocr(paths, noflow, words):
    """run every found image through optical Character Recognition to extract the 
    text the image contains

    https://googlecloudplatform.github.io/google-cloud-python/stable/vision-usage.html#text-detection
    """
    paths = Paths(paths, ImagePath)
    for path in paths:
        f = OCR(path)
        f.scan()
        if words:
            for word in f.words:
                echo.out(word)

        elif noflow:
            echo.out(f.unformatted)

        else:
            echo.out(f.text)


@arg("paths", nargs="+", default=[], help="path to a sound file or directory containing audio files")
@arg("--start", type=Time, default=0, help="transcribe from this time forward (format HH:MM:SS)")
@arg("--stop", type=Time, default=0, help="transcribe to this time (format HH:MM:SS)")
@arg("--language", "--lang", dest="lang", default='en-US', help="The spoken language of the audio file(s)")
def main_speech(paths, start, stop, lang):
    """Create a transcript of each audio file"""
    paths = Paths(paths, SpeechPath)
    for path in paths:
        f = Speech(path, lang)
        f.scan(start, stop)

        for time, text in f:
            echo.out("{:<15}{}", time, text)


# allows for setup.py console scripts
def console():
    exit(__name__)


# allows for "python -m transcribe" syntax
if __name__ == "__main__":
    console()




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import datetime
import re
from datetime import timedelta


class Time(timedelta):
    """Prints time in the format of HH:MM:SS, it will also parse values and normalize
    their values"""
    @property
    def minutes(self):
        return self.seconds // 60

    @property
    def hours(self):
        return self.seconds // 3600

    def __new__(cls, val):
    #def __init__(self, val):
        val = str(val)

        hours = minutes = seconds = 0
        negative = False

        if val:
            if val.startswith("-"):
                val = val[1:]
                negative = True

            if ":" in val:
                parts = list(map(int, val.split(":")))
                parts_count = len(parts)
                if parts_count == 3:
                    hours, minutes, seconds = parts
                elif parts_count == 2:
                    minutes, seconds = parts
                else:
                    raise ValueError("Invalid time {}, try H:MM:SS or NmNs or N".format(val))

            else:
                ms = re.findall(r"(\d+)([hms])", val, re.I)
                if ms:
                    for n, d in ms:
                        d = d.lower()
                        if d == 'h':
                            hours = int(n)
                        elif d == 'm':
                            minutes = int(n)
                        elif d == 's':
                            seconds = int(n)

                else:
                    # https://stackoverflow.com/a/775075/5006
                    if "." in val:
                        val = float(val)
                    m, s = divmod(int(val), 60)
                    h, m = divmod(m, 60)
                    hours = h
                    minutes = m
                    seconds = s

        instance = super(Time, cls).__new__(
            cls,
            hours=hours,
            minutes=minutes,
            seconds=seconds,
        )

        instance.negative = negative
        return instance

    def __str__(self):
        s = super(Time, self).__str__()
        if self.negative:
            s = "-{}".format(s)
        return s

    def total_seconds(self):
        ret = super(Time, self).total_seconds()
        if self.negative:
            ret *= -1
        return ret

    def total_ms(self):
        s = self.total_seconds()
        return s * 1000

    def __format__(self, formatstr):
        return "{{:{}}}".format(formatstr).format(self.__str__())


class String(str):
    @property
    def lines(self):
        return self.splitlines(False)

    def flow(self):
        """This attempts to get rid of extraneous newlines

        NOTE -- this is just really terribly impelemented, it's like 2am and I'm
        tired and I've got no internet for some reason

        :returns: string, the text reflowed
        """
        line_lens = []
        lines = self.lines
        for l in lines:
            line_lens.append(len(l))

        line_lens.sort()
        # we're interested in the longest lines, not the shortest
        if len(line_lens) > 1:
            ret = [""]
            half_i = int(len(line_lens) / 2)
            avg_len = int(sum(line_lens[half_i:]) / half_i)

            #avg_len = int(sum(line_lens) / len(line_lens))
            modifier = 0.4
            min_len = avg_len - int(avg_len * modifier)
            max_len = avg_len + int(avg_len * modifier)
            #pout.v(avg_len, min_len, max_len)

            for l in lines:
                line_len = len(l)
                # ??? -- would it be worth looking at punctuation at the end of the
                # line, if it has it then split, otherwise append
                if line_len >= min_len and line_len <= max_len:
                    ret[-1] += " " + l
                else:
                    ret[-1] += " " + l
                    ret.append("") # new line

            ret = type(self)("\n".join(ret))

        else:
            ret = lines[0]

        return type(self)(ret)

